#define _FILE_OFFSET_BITS 64
#include <stdint.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include "sha2.h"

char fmt_tohex(char c) {
  return (char)(c>=10?c-10+'a':c+'0');
}

size_t fmt_hexdump(char* dest,const unsigned char* s,size_t len) {
  size_t written=0,i;
  if (!dest) return (len>((size_t)-1)/2)?(size_t)-1:len*2;
  for (i=0; i<len; ++i) {
    dest[written]=fmt_tohex(s[i]>>4);
    dest[written+1]=fmt_tohex(s[i]&15);
    written+=2;
  }
  return written;
}

char* SHA512File(const char* filename, char* digest) {
  int fd=open(filename,O_RDONLY);
  char* map;
  off_t ofs=0,flen;
  uint8_t temp[SHA512_DIGEST_LENGTH];
  SHA2_CTX ctx;
  if (fd==-1) return 0;
  flen=lseek(fd,0,SEEK_END);
  SHA512Init(&ctx);
  while (flen) {
    size_t n=(flen>64*1024*1024 ? 64*1024*1024 : flen);
    map=mmap(NULL,flen,PROT_READ,MAP_SHARED,fd,ofs);
    if (map==MAP_FAILED) {
      close(fd);
      return 0;
    }
    SHA512Update(&ctx,(uint8_t*)map,flen);
    munmap(map,flen);
    flen-=n;
    ofs+=n;
  }
  SHA512Final(temp,&ctx);
  close(fd);
  digest[fmt_hexdump(digest,temp,sizeof(temp))]=0;
  return digest;
}

char* SHA256File(const char* filename, char* digest) {
  int fd=open(filename,O_RDONLY);
  char* map;
  off_t ofs=0,flen;
  uint8_t temp[SHA256_DIGEST_LENGTH];
  SHA2_CTX ctx;
  if (fd==-1) return 0;
  flen=lseek(fd,0,SEEK_END);
  SHA256Init(&ctx);
  while (flen) {
    size_t n=(flen>64*1024*1024 ? 64*1024*1024 : flen);
    map=mmap(NULL,flen,PROT_READ,MAP_SHARED,fd,ofs);
    if (map==MAP_FAILED) {
      close(fd);
      return 0;
    }
    SHA256Update(&ctx,(uint8_t*)map,flen);
    munmap(map,flen);
    flen-=n;
    ofs+=n;
  }
  SHA256Final(temp,&ctx);
  close(fd);
  digest[fmt_hexdump(digest,temp,sizeof(temp))]=0;
  return digest;
}
